import os

from mypy_pure.decorators import pure


def impure_leaf() -> None:
    os.remove('file.txt')


def intermediate_2() -> None:
    impure_leaf()


def intermediate_1() -> None:
    intermediate_2()


@pure
def bad() -> None:
    intermediate_1()



def compute_purity(calls: dict[str, set[str]], pure_functions: set[str], blacklist: set[str]) -> dict[str, bool]:
    purity: dict[str, bool] = {}

    def analyze(fn: str, visited: set[str]) -> bool:
        if fn in purity:
            return purity[fn]
        if fn in visited:
            return True  # avoid cycles
        visited.add(fn)

        # Check direct impure calls
        callees = calls.get(fn, set())
        for callee in callees:
            if callee in blacklist or f'builtins.{callee}' in blacklist:
                purity[fn] = False
                return False

        # Check recursive callees
        for callee in callees:
            if callee in pure_functions or callee in calls:
                if not analyze(callee, visited):
                    purity[fn] = False
                    return False

        purity[fn] = True
        return True

    for function in pure_functions:
        analyze(function, set())

    return purity



from __future__ import annotations

import ast


class PurityVisitor(ast.NodeVisitor):
    PURE_DECORATOR_FULLNAME = 'mypy_pure.decorators.pure'

    def __init__(self) -> None:
        self.__imports: dict[str, str] = {}  # alias -> fullname
        self.calls: dict[str, set[str]] = {}  # func_name -> set(callees)
        self.pure_functions: dict[str, int] = {}  # func_name -> lineno
        self.__current_function: str | None = None

    def visit_Import(self, node: ast.Import) -> None:
        for alias in node.names:
            name = alias.asname or alias.name
            self.__imports[name] = alias.name
        self.generic_visit(node)

    def visit_ImportFrom(self, node: ast.ImportFrom) -> None:
        module = node.module or ''
        for alias in node.names:
            name = alias.asname or alias.name
            if module:
                fullname = f'{module}.{alias.name}'
            else:
                fullname = alias.name
            self.__imports[name] = fullname
        self.generic_visit(node)

    def __resolve_name(self, node: ast.AST) -> str | None:
        if isinstance(node, ast.Name):
            return self.__imports.get(node.id, node.id)
        elif isinstance(node, ast.Attribute):
            base = self.__resolve_name(node.value)
            if base:
                return f'{base}.{node.attr}'
        return None

    def visit_FunctionDef(self, node: ast.FunctionDef) -> None:
        # Check for @pure decorator
        is_pure = False
        for decorator in node.decorator_list:
            dec_name = self.__resolve_name(decorator)
            if dec_name == self.PURE_DECORATOR_FULLNAME:
                is_pure = True
            elif isinstance(decorator, ast.Name) and decorator.id == 'pure':
                # Check if 'pure' is imported from the right place
                if self.__imports.get('pure') == self.PURE_DECORATOR_FULLNAME:
                    is_pure = True
                # Or if it's just 'pure' and we assume it's the one (for simple cases)

            elif isinstance(decorator, ast.Attribute) and decorator.attr == 'pure':
                # Handle @decorators.pure
                base = self.__resolve_name(decorator.value)
                if base == 'mypy_pure.decorators':
                    is_pure = True

        if is_pure:
            self.pure_functions[node.name] = node.lineno

        prev_function = self.__current_function
        self.__current_function = node.name
        self.calls[node.name] = set()

        self.generic_visit(node)

        self.__current_function = prev_function

    def visit_Call(self, node: ast.Call) -> None:
        if self.__current_function is None:
            self.generic_visit(node)
            return

        callee_name = None
        if isinstance(node.func, ast.Name):
            callee_name = self.__imports.get(node.func.id, node.func.id)
        elif isinstance(node.func, ast.Attribute):
            base = self.__resolve_name(node.func.value)
            if base:
                callee_name = f'{base}.{node.func.attr}'

        if callee_name:
            self.calls[self.__current_function].add(callee_name)

        self.generic_visit(node)



from typing import Callable, ParamSpec, TypeVar, cast

P = ParamSpec('P')
R = TypeVar('R')


def pure(func: Callable[P, R]) -> Callable[P, R]:
    setattr(func, '__pure__', True)
    return cast(Callable[P, R], func)



from __future__ import annotations

import ast
import configparser
import sys

from mypy.nodes import MypyFile
from mypy.options import Options
from mypy.plugin import Plugin

from mypy_pure.configuration import BLACKLIST
from mypy_pure.purity.analyzer import compute_purity
from mypy_pure.purity.visitor import PurityVisitor


class PurityPlugin(Plugin):
    def __init__(self, options: Options) -> None:
        super().__init__(options)
        self.__checked_files: set[str] = set()
        self.__blacklist = BLACKLIST.copy()
        self.__load_config(options)

    def __load_config(self, options: Options) -> None:
        if not options.config_file:
            return

        config = configparser.ConfigParser()
        try:
            config.read(options.config_file)
            if 'mypy-pure' in config:
                impure_funcs = config['mypy-pure'].get('impure_functions', '')
                for func in impure_funcs.split(','):
                    func = func.strip()
                    if func:
                        self.__blacklist.add(func)
        except Exception:
            pass

    def get_additional_deps(self, file: MypyFile) -> list[tuple[int, str, int]]:
        if file.fullname in self.__checked_files:
            return []

        # Skip stdlib and other system modules to avoid noise and performance hit
        # This list is heuristic.
        if file.fullname.startswith(('builtins', 'typing', 'sys', 'os', 'abc', 'enum', 'mypy.', '_')):
            return []

        self.__checked_files.add(file.fullname)

        try:
            # We need to read the source file again because MypyFile doesn't expose the raw source easily here,
            # and we want to parse it with ast.
            if not file.path:
                return []

            with open(file.path, 'r', encoding='utf-8') as f:
                source = f.read()

            tree = ast.parse(source, filename=file.path)
            visitor = PurityVisitor()
            visitor.visit(tree)

            if not visitor.pure_functions:
                return []

            purity_map = compute_purity(visitor.calls, set(visitor.pure_functions.keys()), self.__blacklist)

            for fn, lineno in visitor.pure_functions.items():
                if not purity_map.get(fn, True):
                    # Report error in a format mypy/editors understand
                    msg = (
                        f'{file.path}:{lineno}: '
                        f"error: Function '{fn}' is annotated as pure but calls impure functions.\n"
                    )
                    sys.stdout.write(msg)
                    sys.stdout.flush()

        except Exception as _exc:  # noqa
            # Silently fail or print debug info if needed
            # sys.stdout.write(f'AST Analysis failed for {file.fullname}: {e}\n')
            pass

        return []


def plugin(version: str) -> type[PurityPlugin]:
    return PurityPlugin



BLACKLIST = {
    # Builtins
    'builtins.open',
    'builtins.print',
    'builtins.input',
    # os module - File/Dir operations
    'os.remove',
    'os.unlink',
    'os.rmdir',
    'os.mkdir',
    'os.makedirs',
    'os.rename',
    'os.replace',
    'os.link',
    'os.symlink',
    'os.chmod',
    'os.chown',
    'os.utime',
    'os.truncate',
    'os.ftruncate',
    # os module - Process/System
    'os.system',
    'os.popen',
    'os.spawnl',
    'os.spawnle',
    'os.spawnlp',
    'os.spawnlpe',
    'os.spawnv',
    'os.spawnve',
    'os.spawnvp',
    'os.spawnvpe',
    'os.execl',
    'os.execle',
    'os.execlp',
    'os.execlpe',
    'os.execv',
    'os.execve',
    'os.execvp',
    'os.execvpe',
    'os.fork',
    'os.kill',
    'os.abort',
    'os.chdir',
    'os.fchdir',
    # os module - I/O
    'os.open',
    'os.read',
    'os.write',
    'os.close',
    'os.pipe',
    # sys module
    'sys.exit',
    'sys.stdout.write',
    'sys.stderr.write',
    # subprocess
    'subprocess.run',
    'subprocess.call',
    'subprocess.check_call',
    'subprocess.check_output',
    'subprocess.Popen',
    # shutil
    'shutil.rmtree',
    'shutil.copy',
    'shutil.copy2',
    'shutil.copyfile',
    'shutil.copytree',
    'shutil.move',
    'shutil.chown',
    # time
    'time.sleep',
    # random (state modification)
    'random.seed',
    # socket
    'socket.socket',
    'socket.create_connection',
    # pathlib (common write ops)
    'pathlib.Path.write_text',
    'pathlib.Path.write_bytes',
    'pathlib.Path.touch',
    'pathlib.Path.mkdir',
    'pathlib.Path.unlink',
    'pathlib.Path.rmdir',
    'pathlib.Path.rename',
    'pathlib.Path.replace',
}



import unittest

from mypy_pure.decorators import pure


class TestPureDecorators(unittest.TestCase):
    def test_pure_sets_attribute(self) -> None:
        @pure
        def add(a, b):
            return a + b

        self.assertTrue(getattr(add, '__pure__', False))
        self.assertEqual(add(2, 3), 5)



from __future__ import annotations

from pathlib import Path
from unittest import TestCase


class TestPlugin(TestCase):
    def __run_mypy(self, file_path: Path, config_file: Path | None = None) -> tuple[str, str, int]:
        tests_path = Path(__file__).resolve().parent
        if config_file is None:
            config_file = tests_path / 'resources' / 'mypy.ini'

        import sys
        from io import StringIO

        from mypy.api import run as mypy_api_run

        capture = StringIO()
        old_stdout = sys.stdout
        sys.stdout = capture
        try:
            stdout, stderr, exit_status = mypy_api_run(
                [
                    '--config-file',
                    str(config_file),
                    '--no-error-summary',
                    '--hide-error-context',
                    '--follow-imports=silent',
                    '--strict',
                    '--no-incremental',
                    str(file_path),
                ]
            )
            # Combine captured stdout with mypy's returned stdout
            combined_stdout = stdout + capture.getvalue()
            return combined_stdout, stderr, exit_status
        finally:
            sys.stdout = old_stdout

    def _get_resource_path(self, filename: str) -> Path:
        return Path(__file__).resolve().parent / 'resources' / filename

    def test_pure_function_calls_custom_impure(self):
        resource = self._get_resource_path('pure_calls_custom_impure.py')
        config = self._get_resource_path('mypy_custom.ini')
        stdout, stderr, exit_status = self.__run_mypy(resource, config)
        self.assertEqual(0, exit_status)
        self.assertIn(
            'is annotated as pure but calls impure functions',
            stdout,
            f'Expected purity violation, got: {stdout}',
        )

    def test_simple_pure_function(self):
        resource = self._get_resource_path('pure_is_ok.py')
        stdout, stderr, exit_status = self.__run_mypy(resource)
        self.assertEqual(0, exit_status)
        self.assertEqual(stdout.strip(), '', f'Unexpected mypy output: {stdout}')

    def test_pure_function_calls_impure_stdlib(self):
        resource = self._get_resource_path('pure_calls_impure_stdlib.py')
        stdout, stderr, exit_status = self.__run_mypy(resource)
        self.assertEqual(0, exit_status)
        self.assertIn(
            'is annotated as pure but calls impure functions',
            stdout,
            f'Expected purity violation, got: {stdout}',
        )

    def test_pure_function_calls_impure_function_indirectly(self):
        resource = self._get_resource_path('pure_calls_impure_indirect.py')
        stdout, stderr, exit_status = self.__run_mypy(resource)
        self.assertEqual(0, exit_status)
        self.assertIn(
            'is annotated as pure but calls impure functions',
            stdout,
            f'Expected purity violation, got: {stdout}',
        )

    def test_pure_function_calls_impure_function_deeply_indirect(self):
        resource = self._get_resource_path('pure_calls_impure_deeply_indirect.py')
        stdout, stderr, exit_status = self.__run_mypy(resource)
        self.assertEqual(0, exit_status)
        self.assertIn(
            'is annotated as pure but calls impure functions',
            stdout,
            f'Expected purity violation, got: {stdout}',
        )

    def test_pure_function_calls_print(self):
        resource = self._get_resource_path('pure_calls_print.py')
        stdout, stderr, exit_status = self.__run_mypy(resource)
        self.assertEqual(0, exit_status)
        self.assertIn(
            'is annotated as pure but calls impure functions',
            stdout,
            f'Expected purity violation, got: {stdout}',
        )

    def test_pure_function_calls_sleep(self):
        resource = self._get_resource_path('pure_calls_sleep.py')
        stdout, stderr, exit_status = self.__run_mypy(resource)
        self.assertEqual(0, exit_status)
        self.assertIn(
            'is annotated as pure but calls impure functions',
            stdout,
            f'Expected purity violation, got: {stdout}',
        )

    def test_pure_function_calls_pure_function(self):
        resource = self._get_resource_path('pure_calls_pure.py')
        stdout, stderr, exit_status = self.__run_mypy(resource)
        self.assertEqual(0, exit_status)
        self.assertEqual(stdout.strip(), '', f'Unexpected mypy output: {stdout}')
